/**
 * Judy
 * Valladolid-like programming problem judge.
 *
 * Written by Igor Naverniouk
 * igor@lexansoft.com
 *
 * version 0.9beta
 * 2003
 *
 * USAGE:
 *     1. Edit judy.cfg to your liking
 *     2. Add entries to teams.txt and probs.txt
 *     3. Run ./judy judy.cfg (starts the judy daemon (judyd))
 *     4. Configure the CGI script (submit.pl) to make it aware of judyd's port number.
 *
 * NOTES:
 *     1. As far as Judy is concerned, the log file is everything, i.e.
 *       "If It Ain't In The Log, It Didn't Happen".
 *     2. Run Judy only as a separate user, never as yourself, and
 *        definitely never as root!
 *     3. Judy0.9 is written in C with C++ conveniences.
 **/

#include <stdio.h>
#include <fstream>
#include <map>
#include <string>
#include <sstream>
#include <iostream>
#include <ctype.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/times.h>
#include <sys/wait.h>
#include <sys/ptrace.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>
#include "log.h"
#include "db.h"

using namespace std;

/** BEGIN GLOBALS **/
#define RECOMPILES 100       // Number of times to recompile when gcc is crashing

/**
 * Contains key-value configuration variable pairs,
 * a la Java's Properties class
 **/
map< string, string > g_config;

/** END GLOBALS **/

/**
 * Reads the config file and sets thigs up
 **/
void init( char *file )
{
    ifstream fin( file, ios::in );
    string line;
    while( getline( fin, line ) )
    {
        if( !line.size() || line[0] == '#' ) continue;
        string key, value;
        istringstream sin( line );
        if( sin >> key >> value ) g_config[key] = value;
    }
    fin.close();

    initlog( g_config["logfile"], g_config["weblog"] );
    initdb( g_config["teamfile"], g_config["probfile"], g_config["logfile"] );

    updateHTML();
}

/**
 * Brutally kills all processes except for some harmless ones
 * in their name.
 *
 * FIXME: Stuff is hard-coded and a specific ps output format is assumed
 **/
void cleanProcesses()
{
    FILE *pin = popen( "ps -ucs448", "r" );
    if( pin )
    {
        char c;
        while( fscanf( pin, "%c", &c ) && c != '\n' );  // Skip the first line
        int pid;
        char buf[64];
        while( fscanf( pin, " %d %s %s %s ", &pid, buf, buf, buf ) )
        {
            // Keep several processes alive
            if( !strcmp( buf, "bash"  ) ||
                !strcmp( buf, "ps"    ) ||
                !strcmp( buf, "judyd" ) ) continue;

            // Kill the rest
            if( pid != getpid() ) kill( pid, 15 );
        }
        pclose( pin );
    }
}

/**
 * Checks the ccerr file in the execute directory and
 * returns true if it contains gcc's BS about "internal
 * compiler error" in gcc.c:2328
 **/
bool gccbug()
{
    char file[1024];
    sprintf( file, "%s/ccerr", g_config["rundir"].c_str() );
    ifstream fin( file, ios::in );
    bool found = false;
    string line;
    while( getline( fin, line ) )
        if( strstr( line.c_str(), "gcc.c:2328" ) ||
            strstr( line.c_str(), "Internal compiler error" ) )
            found = true;
    fin.close();
    return found;
}

/**
 * Returns the time in milliseconds
 **/
long long gettime()
{
    struct tms buf;
    static const long CPS = sysconf( _SC_CLK_TCK );
    return ( long long )times( &buf ) * 1000 / CPS;
}

/**
 * Invoked by the "Submit-o-matic"-like Perl script.
 * Parameters:
 *     - ConfigFile
 *     - TeamID
 *     - ProblemID
 *     - Language
 *     - CodeFile
 **/
int main( int argc, char **argv )
{
    // To prevent fork()ed process orphanage (<defunct> business)
//    signal( SIGCHLD, SIG_IGN );

    init( argv[1] );

    if( argc == 2 )
    {
        // Initialization: start the Judy daemon
        char judyd[1024];
        sprintf( judyd, "./judyd %s %s %s %s",
            g_config["incdir"].c_str(),
            g_config["pidfile"].c_str(),
            g_config["port"].c_str(),
            argv[1] );
        int err;
        if( ( err = system( judyd ) ) )
            printf( "Could not start judyd; error %d\n", err / 256 );
        return 0;
    }

    if( argc != 6 ) return ERROR_ARGC;

    // Beautify input
    for( int i = 0; argv[3][i]; i++ ) argv[3][i] = tolower( argv[3][i] );
    for( int i = 0; argv[4][i]; i++ ) argv[4][i] = tolower( argv[4][i] );

    if( !teamExists( argv[2] ) )
    {
        log( argv[2], "?", "?", ERROR_TEAM );
        return ERROR_TEAM;
    }

    if( !probExists( argv[3] ) )
    {
        log( argv[2], argv[3], "?", ERROR_PROB );
        return ERROR_PROB;
    }

    if( !langExists( argv[4] ) )
    {
        log( argv[2], argv[3], argv[4], ERROR_LANG );
        return ERROR_LANG;
    }

    // Figure out the code file name
    int i = strlen( argv[5] );
    while( i + 1 && argv[5][i] != '/' ) i--;

    char cmd[1024];

    // Clean the run directory
    sprintf( cmd, "rm -f %s/*", g_config["rundir"].c_str() );
    system( cmd );

    // Copy code to run directory
    sprintf( cmd, "cp %s %s/%s.%s",
        argv[5],                        // from
        g_config["rundir"].c_str(),
        argv[3],                        // Source file base name
        sourceSuffix( argv[4] ) );      // Source file suffix
    if( system( cmd ) )
    {
        log( argv[2], argv[3], argv[4], ERROR_COPY );
        return ERROR_COPY;
    }

    sleep(1);    // A pathetic attempt to prevent gcc crash (doesn't seem to work)

    // Compile
    sprintf( cmd, "cd %s; %s %s.%s 2> ccerr",
        g_config["rundir"].c_str(),
        compileCmd( argv[4] ),
        argv[3],
        sourceSuffix( argv[4] ) );
    fclose( stderr );
    int compileCount = RECOMPILES;
    bool compErr;
    while( ( compErr = system( cmd ) ) && gccbug() && compileCount-- );
    if( compErr )
    {
        // Save ccerr for the "details" link
        sprintf( cmd, "cp %s/ccerr %s/errors/%d.txt",
            g_config["rundir"].c_str(),
            g_config["weblog"].c_str(),
            getRunID() );
        system( cmd );

        log( argv[2], argv[3], argv[4], ERROR_COMP );
        return ERROR_COMP;
    }

    // Execute
    sprintf( cmd, "%s/%s < %s/%s.in > %s/out",
        g_config["rundir"].c_str(),
        executeCmd( argv[4] ),
        g_config["iodir"].c_str(),
        argv[3],
        g_config["rundir"].c_str() );

    int pid = fork();
    if( pid == -1 )
    {
        cleanProcesses();
        log( argv[2], argv[3], argv[4], ERROR_FORK );
        return ERROR_FORK;
    }
    else if( !pid )
    {
        // CHILD: Run user program
        ptrace( PTRACE_TRACEME, 0, 0, 0 );
        execl( "/bin/sh", "/bin/sh", "-c", cmd, NULL );
        return -1;                   // Should never get here
    }
    else
    {
        // PARENT: Time the running program and catch signals.
        long long startTime = gettime();
        int pid2 = fork();
        if( pid2 == -1 )
        {
            cleanProcesses();
            log( argv[2], argv[3], argv[4], ERROR_FORK );
            return ERROR_FORK;
        }
        else if( !pid2 )
        {
            // CHILD2: Kill the child after time limit expires
            sleep( 31 );                // DEBUG: Hardcoded 30+1 seconds
            kill( pid, SIGTERM );       // Errors here don't matter
            return 0;
        }
        else
        {
            // PARENT: Catch child's signals
            int status;
            while( true )
            {
                wait( &status );
                if( gettime() - startTime >= 30000LL )      // DEBUG: Hardcoded 30000 milliseconds
                {
                    // Time limit exceeded
                    ptrace( PTRACE_KILL, pid, 0, 0 );       // Errors here don't matter
                    kill( pid2, SIGTERM );
                    log( argv[2], argv[3], argv[4], ERROR_TLEX );
                    return ERROR_TLEX;
                }
                if( status != 1407 )
                {
                    int sig = status / 256;
                    if( sig == SIGBUS ) sig = 11;           // Treat bus errors as seg faults
                    if( sig == 3 || sig == 4 || sig == 5 || sig == 6 || sig == 8 || sig == 11 )
                    {
                        // Child died with a signal 'sig'
                        kill( pid2, SIGTERM );
                        log( argv[2], argv[3], argv[4], ERROR_SIGL + sig * 256 );
                        return ERROR_SIGL;
                    }
                    break;  // User program exited normally
                }
                if( ptrace( PTRACE_CONT, pid, 0, 0 ) )
                {
                    kill( pid2, SIGTERM );
                    log( argv[2], argv[3], argv[4], ERROR_PTRC );
                    return ERROR_PTRC;
                }
            }
            ptrace( PTRACE_KILL, pid, 0, 0 );
            kill( pid2, SIGTERM );
        }
    }

    // Compare output
    sprintf( cmd, "diff %s/out %s/%s.out",
        g_config["rundir"].c_str(),
        g_config["iodir"].c_str(),
        argv[3] );
    if( system( cmd ) )
    {
        log( argv[2], argv[3], argv[4], ERROR_DIFF );
        return ERROR_DIFF;
    }

    // Remove the output file!!!
    sprintf( cmd, "rm -f %s/out", g_config["rundir"].c_str() );
    system( cmd );

    // Success
    log( argv[2], argv[3], argv[4], 0 );
    return 0;
}
