/**
 * Deals with logging and HTML generation
 * The log is a sequence of lines in chronological order. Each line is
 *     <runID> <user> <problem> <language> <reply>
 * The <runID> is unique for each line and increases by 1 every time.
 **/

#include <fstream>
#include <string>
#include <deque>
#include <set>
#include <sstream>
#include "db.h"
#include "log.h"

using namespace std;

static const char SIGNALS[][16] =
{
    "", "", "", "SIGQUIT", "SIGILL", "SIGTRAP", "SIGABRT", "", "SIGFPE",
    "", "", "SIGSEGV"
};

/**
 * Log and HTML files stay closed between loggings
 **/
static string logfile;
static string weblogd;

static int nextRunID;

/**
 * Returns the run ID of the current run (next to be logged)
 **/
int getRunID()
{
    return nextRunID;
}

/**
 * Sets the log file
 **/
void initlog( string file, string weblogdir )
{
    logfile = file;
    weblogd = weblogdir;

    // Figure out what the next run ID should be
    nextRunID = 0;
    ifstream in( logfile.c_str(), ios::in );
    string line;
    while( in >> nextRunID ) getline( in, line );
    nextRunID++;
    in.close();
}

/** Appends a line to log file and updates the web page **/
void log( const char *team, const char *prob, const char *lang, int status )
{
    ofstream fout( logfile.c_str(), ios::app );
    fout << nextRunID++ << " "
         << team << " "
         << prob << " "
         << lang << " ";
    if( status ) fout << "no " << status;
    else fout << "yes";
    fout << endl;
    fout.close();

    updateHTML();
}

/** Updates judy.html (HTML log) and stats.html (scoreboard) **/
void updateHTML()
{
    ifstream in( logfile.c_str(), ios::in );
    ofstream out( ( weblogd + "/judy.html" ).c_str(), ios::out );

    // Keep a record of which user solved which problem
    set< pair< string, string > > solved;
    set< pair< string, string > > tried;

    int id;
    string team, prob, lang, yesno;

    // Get the last WLOGSIZE entries from the log
    deque< string > lines;
    string line;
    while( getline( in, line ) && line.size() )
    {
        // Update stats.html
        istringstream linein( line );
        linein >> id >> team >> prob >> lang >> yesno;
        if( yesno == "yes" )
            solved.insert( pair< string, string >( team, prob ) );
        else if( yesno == "no" && team != "?" && prob != "?" )
            tried.insert( pair< string, string >( team, prob ) );

        lines.push_back( line );
        if( lines.size() > WLOGSIZE ) lines.pop_front();
    }

    out << "<html><head><title>CPSC448: Real-time Status</title></head>\n";
    out << "<body><table width=640 border=0><tr><td>\n\n";
    out << "<h2>CPSC448: Submissions log</h3>\n";
    out << "<p>Assignment 1 (Problems A, B, C and D) is past due now.</p>";
    out << "<table border=0><tr><td bgcolor=#FFFFCC>System upgrade in progress... You may get strange/incorrect replies!</td></tr></table><br>\n";
    out << "<table border=0><tr><td bgcolor=#FFCCCC>Problem G description is wrong! Output a blank line after <u>every</u> test case, not just between the test cases.</td></tr></table><br>\n";
    out << "<table border=0><tr><td bgcolor=#FFCCCC>Please wait for a reply before submitting again. You can get a Wrong Answer if you don't wait.</td></tr></table><br>\n";
    out << "<table border=1>\n";
    out << "    <tr>\n";
    out << "        <td><b>RunID</b></td>\n";
    out << "        <td><b>User</b></td>\n";
    out << "        <td><b>Problem</b></td>\n";
    out << "        <td><b>Language</b></td>\n";
    out << "        <td><b>Reply</b></td>\n";
    out << "    </tr>\n";

    for( deque< string >::reverse_iterator it = lines.rbegin(); it != lines.rend(); it++ )
    {
        istringstream linein( *it );
        linein >> id >> team >> prob >> lang >> yesno;
        string reply;
        if( yesno == "yes" ) reply = "<font color=#00CC00>Correct</font>";
        else
        {
            int status; linein >> status;
            ostringstream ostr;
            switch( status )
            {
                case ERROR_ARGC:
                    reply = "Internal error of the first kind";
                    break;
                case ERROR_TEAM:
                    reply = "<font color=#0000CC>I don't recognize you, ";
                    reply += team;
                    reply += "</font>";
                    break;
                case ERROR_PROB:
                    reply = "<font color=#0000CC>Solving a problem that doesn't exist?</font>";
                    break;
                case ERROR_LANG:
                    reply = "<font color=#0000CC>Sorry, I don't speak ";
                    reply += lang;
                    reply += "</font>";
                    break;
                case ERROR_COPY:
                    reply = "Internal error of the second kind";
                    break;
                case ERROR_COMP:
                    ostr << "<font color=#0000CC>Compile error</font>"
                         << "(<a href=\"errors/" << id << ".txt\">details</a>)";
                    reply = ostr.str();
                    break;
                case ERROR_FORK:
                    reply = "Internal error of the third kind";
                    break;
                case ERROR_DIFF:
                    reply = "<font color=#CC0000>Wrong answer</font>";
                    break;
                case ERROR_PTRC:
                    reply = "Internal error of the fourth kind";
                    break;
                case ERROR_TLEX:
                    reply = "<font color=#CC0000>Time limit exceeded</font>";
                    break;
                default:        // Died with a signal
                    if( status % 256 != ERROR_SIGL )
                        reply = "Big time internal screw-up!";
                    else
                    {
                        reply = "<font color=#CC0000>Runtime error (";
                        reply += SIGNALS[status / 256];
                        reply += ")</font>";
                    }
                    break;
            }
        }
        out << "    <tr bgcolor=#EEEEEE>\n";
        out << "        <td>" << id << "</td>\n";
        out << "        <td>" << team << "</td>\n";
        out << "        <td>" << prob << "</td>\n";
        out << "        <td>" << lang << "</td>\n";
        out << "        <td>" << reply << "</td>\n";
        out << "    </tr>\n";
    }

    out << "</table>\n\n";

    out << "</td></tr></table>\n</body>\n</html>\n";

    out.close();
    in.close();

    // Now update stats.html
    ofstream out2( ( weblogd + "/stats.html" ).c_str(), ios::out );

    out2 << "<html><head><title>CPSC448: Submission Statistics by user</title></head>\n";
    out2 << "<body><table width=640 border=0><tr><td>\n\n";
    out2 << "<h2>CPSC448: Submission Statistics by user</h3>\n";
    out2 << "<table border=1>\n";
    out2 << "    <tr>\n";
    out2 << "        <td></td>\n";
    for( set< string >::iterator i = g_probs.begin(); i != g_probs.end(); i++ )
        out2 << "        <td><b>" << *i << "</b></td>\n";
    out2 << "    </tr>\n";
    for( map< string, Team >::iterator t = g_teams.begin(); t != g_teams.end(); t++ )
    {
        out2 << "    <tr>\n";
        out2 << "        <td>" << t->first << "</td>\n";
        for( set< string >::iterator i = g_probs.begin(); i != g_probs.end(); i++ )
        {
            pair< string, string > p( t->first, *i );
            out2 << "        <td width=40 bgcolor=";
            if( solved.count( p ) ) out2 << "#DDFFDD>Yes";
            else if( tried.count( p ) ) out2 << "#FFFFDD>Tried";
            else out2 << "#FFDDDD>No";
            out2 << "</td>\n";
        }
        out2 << "    </tr>\n";
    }
    out2 << "</table>\n";
    out2 << "</td></tr></table></body></html>\n";
    out2.close();
}

void updateStats()
{
}
